@@ -13,16 +13,19 @@ For python startup when an ancestor process has set the env indicating
 that code coverage is being collected we activate coverage based on
 info passed via env vars.
 """
+import atexit
 import os
 import signal
 
-active_cov = None
+_active_cov = None
 
 
 def multiprocessing_start(_):
+    global _active_cov
     cov = init()
     if cov:
-        multiprocessing.util.Finalize(None, cleanup, args=(cov,), exitpriority=1000)
+        _active_cov = cov
+        multiprocessing.util.Finalize(None, cleanup, exitpriority=1000)
 
 
 try:
@@ -36,7 +39,7 @@ else:
 def init():
     # Only continue if ancestor process has set everything needed in
     # the env.
-    global active_cov
+    global _active_cov
 
     cov_source = os.environ.get('COV_CORE_SOURCE')
     cov_config = os.environ.get('COV_CORE_CONFIG')
@@ -44,11 +47,13 @@ def init():
     cov_branch = True if os.environ.get('COV_CORE_BRANCH') == 'enabled' else None
 
     if cov_datafile:
+        if _active_cov:
+            cleanup()
         # Import what we need to activate coverage.
         import coverage
 
         # Determine all source roots.
-        if cov_source == os.pathsep:
+        if cov_source in os.pathsep:
             cov_source = None
         else:
             cov_source = cov_source.split(os.pathsep)
@@ -56,7 +61,7 @@ def init():
             cov_config = True
 
         # Activate coverage for this process.
-        cov = active_cov = coverage.coverage(
+        cov = _active_cov = coverage.Coverage(
             source=cov_source,
             branch=cov_branch,
             data_suffix=True,
@@ -75,23 +80,57 @@ def _cleanup(cov):
     if cov is not None:
         cov.stop()
         cov.save()
+        cov._auto_save = False  # prevent autosaving from cov._atexit in case the interpreter lacks atexit.unregister
+        try:
+            atexit.unregister(cov._atexit)
+        except Exception:
+            pass
 
 
-def cleanup(cov=None):
-    global active_cov
+def cleanup():
+    global _active_cov
+    global _cleanup_in_progress
+    global _pending_signal
 
-    _cleanup(cov)
-    if active_cov is not cov:
-        _cleanup(active_cov)
-    active_cov = None
+    _cleanup_in_progress = True
+    _cleanup(_active_cov)
+    _active_cov = None
+    _cleanup_in_progress = False
+    if _pending_signal:
+        _signal_cleanup_handler(*_pending_signal)
+        _pending_signal = None
 
 
 multiprocessing_finish = cleanup  # in case someone dared to use this internal
 
+_previous_handlers = {}
+_pending_signal = None
+_cleanup_in_progress = False
 
-def _sigterm_handler(*_):
+
+def _signal_cleanup_handler(signum, frame):
+    global _pending_signal
+    if _cleanup_in_progress:
+        _pending_signal = signum, frame
+        return
     cleanup()
+    _previous_handler = _previous_handlers.get(signum)
+    if _previous_handler == signal.SIG_IGN:
+        return
+    elif _previous_handler and _previous_handler is not _signal_cleanup_handler:
+        _previous_handler(signum, frame)
+    elif signum == signal.SIGTERM:
+        os._exit(128 + signum)
+    elif signum == signal.SIGINT:
+        raise KeyboardInterrupt()
+
+
+def cleanup_on_signal(signum):
+    previous = signal.getsignal(signum)
+    if previous is not _signal_cleanup_handler:
+        _previous_handlers[signum] = previous
+        signal.signal(signum, _signal_cleanup_handler)
 
 
 def cleanup_on_sigterm():
-    signal.signal(signal.SIGTERM, _sigterm_handler)
+    cleanup_on_signal(signal.SIGTERM)
